{% extends "base.html" %}
{% load leaflet_storage_tags %}
{% block title %}{{ map.name }}{% endblock %}
{% block body_class %}map_detail{% endblock %}

{% block extra_head %}
    {{ block.super }}
    {% leaflet_storage_css %}
    {% leaflet_storage_js locale=locale %}
    {% if object.share_status != object.PUBLIC %}
        <meta name="robots" content="noindex">
    {% endif %}
{% endblock %}

{% block content %}
    {% block map_init %}
        {% include "leaflet_storage/map_init.html" %}
    {% endblock %}
    {% include "leaflet_storage/map_messages.html" %}
{% endblock %}
