{% load i18n %}
<h3>{% trans "Map permissions" %}</h3>
<form action="{% url 'map_update_permissions' map.pk %}" method="post" id="map_edit">
    {% csrf_token %}
    {{ form }}
    <input type="submit" />
</form>