"Create collaborative maps on top of Geodjango and Leaflet."
VERSION = (0, 8, 0, 'beta', 0)

__author__ = 'Yohan Boniface'
__contact__ = "yb@enix.fr"
__homepage__ = "https://github.com/umap-project/django-leaflet-storage"
__version__ = ".".join(map(str, VERSION))
